/**
 * tyrre - small "web-server" for dyndns applications
 * tyrre must be run through tcpserver from DJB's ucspi-tcp suite
 * compile with: diet gcc -Os -Wall -o tyrre tyrre.c; sstrip tyrre
 * Frank Bergmann 07/2009
 */

/*
 * changes:
 * 0.12:   - add short URL feature
 * 0.11:   - output buffered
 *         - write successful dyndns registration to stderr
 *         - use client IP 0.0.0.0 if TCPREMOTEIP unset
 * 0.10:   - recognize X-Forwarded-For tag
 * 0.9:    - use string funcs of lib due to fast and small (asm) funcs in dietlibc
 * 0.8:    - write http answer as one output string to prevent tcpserver
 *           splitting this response into two ip packets
 *         - return IP# even if dyndns file couldn't be written, send
 *           also string "Error" in this case
 * 0.7:    - prefer output to net over output to file
 *         - use '\r\n' line termination
 * 0.6:    logging output to stderr
 */

#include <stdint.h>
#include <stdbool.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
 
#define BUFSIZE 8192

#define HEADER200 "HTTP/1.0 200 OK\r\nServer: tyrre/0.12\r\nContent-Type: text/plain\r\n\r\n"

static const char str200[] = HEADER200 "@@@.@@@.@@@.@@@\n";
static const char str301[] = "HTTP/1.0 301 Moved Permanently\r\nLocation: ";
static const char str302[] = "HTTP/1.0 302 Found\r\nLocation: http://www.tuxad.com/\r\nContent-Type: text/html\r\n\r\n<title>302 Found</title>\r\nThe document is <a href=\"http://www.tuxad.com/\">here</a>.";
static const char logdynstr[] = "dyndns: ";

#include "tyrre-dyndnstable.c"
/**
 * dyndnstable sample:
#define TABLE_LEN 1
static char* map_table[TABLE_LEN][4] = {
  { "key1_xyz", "", "user1.mydomain.de", "\n120\n" }
};
 *
 */

#define CRC_ALGO_BIT_BY_BIT_FAST 1
typedef uint32_t crc_t;



/**
 * \file pycrc_stdout
 * Functions and types for CRC checks.
 *
 * Generated on Sun Nov 13 14:58:19 2011,
 * by pycrc v0.7.8, http://www.tty1.net/pycrc/
 * using the configuration:
 *    Width        = 30
 *    Poly         = 0x2030b9c7
 *    XorIn        = 0x00000000
 *    ReflectIn    = False
 *    XorOut       = 0x00000000
 *    ReflectOut   = False
 *    Algorithm    = bit-by-bit-fast
 *
 * python pycrc.py --algorithm=bit-by-bit-fast --width 30 --poly 0x2030B9C7 --reflect-in False --xor-in 0x0 --reflect-out False --xor-out 0x000000 --generate=c-main >crc30.c
 *
 *****************************************************************************/

/**
 * Update the crc value with new data.
 *
 * \param crc      The current crc value.
 * \param data     Pointer to a buffer of \a data_len bytes.
 * \param data_len Number of bytes in the \a data buffer.
 * \return         The updated crc value.
 *****************************************************************************/
static crc_t crc_update(crc_t crc, const unsigned char *data, size_t data_len) {
  unsigned int i;
  bool bit;
  unsigned char c;

  while (data_len--) {
    c = *data++;
    for (i = 0x80; i > 0; i >>= 1) {
      bit = crc & 0x20000000;
      if (c & i) {
        bit = !bit;
      }
      crc <<= 1;
      if (bit) {
        crc ^= 0x2030b9c7;
      }
    }
    crc &= 0x3fffffff;
  }
  return crc & 0x3fffffff;
}

/* for short URL */
static unsigned char asciitrans(uint8_t num) {
  if (num < 10) return num + 48;
  if (num < 36) return num + 55;
  if (num < 62) return num + 61;
  if (num < 63) return '%';
  return '_';
}

static char *convert_url(register char *u) {
  char *cptr = u;

  /* convert %2F and %3F */
  for (;*u;++u) {

    if (*u != '%') continue;
    ++u;
    if (*u == '3') goto test3f;
    if (*u != '2') continue;
    ++u;
    if (*u != 'F') continue;
    u -= 2;
    *u = '/';
    goto shift;

    test3f:
    ++u;
    if (*u != 'F') continue;
    u -= 2;
    *u = '?';

    shift:
    {
      int c;
      for (c=1;u[c+2];++c) u[c] = u[c+2];
      u[c] = 0;
    }
  }

  /* starts with http:// ? ignore it! */
  if ((cptr[0] == 'h') && (cptr[1] == 't') && (cptr[2] == 't') && (cptr[3] == 'p')) {
    if ((cptr[4] == ':') && (cptr[5] == '/') && (cptr[6] == '/')) {
      cptr += 7;
    } else {
      if ((cptr[4] == '%') && (cptr[5] == '3')) {
        if ((cptr[6] == 'A') || (cptr[6] == 'a')) {
          if ((cptr[7] == '/') && (cptr[8] == '/'))
            cptr += 9;
        }
      }
    }
  }
  return cptr;
}

void my_exit(const int st) { write1(str302, sizeof str302 - 1); _exit(st); }

int main() {
  register char *cptr, *u;
  register char *p, *a;
  register ssize_t i;
  char ibuf[BUFSIZE + 1];
  char ur[BUFSIZE + 1];
  char outbuf[BUFSIZE + 1];

  cptr = ibuf; u = ur;
  /**
   * we don't watch for a complete header!
   * just the first line will be parsed (and maybe only read)
   * we don't want to work with signals and timeouts (overhead here)
   */
  i = read(0, cptr, BUFSIZE);
  if (i < 15) _exit(1); /* data too short for HTTP-request, don't answer it */
  cptr[i] = 0;

  if (i == BUFSIZE) {
    /* throw away more incoming data */
    while ((i = read(0, u, BUFSIZE)) == BUFSIZE);
  }

  /* start page will be immediately redirected */
  if (!strncmp(cptr, "GET / HTTP/1.", 13)) my_exit(0); /* ok, quick redirect answer */
  if (strncmp(cptr, "GET /", 5)) _exit(2); /* not a HTTP-request, don't answer */

  /* cptr points to (relative) URL */
  cptr += 5;

  u = strchr(cptr, ' '); /* find end of URL-string */
  if (!u) _exit(2);

  /* check if it contains "HTTP/1." after "GET /.... " */
  if (strncmp(u, " HTTP/1.", 8)) my_exit(4); /* error, but gracefully answer */

  *u = 0;
  ++u;
  u = strstr(u, "X-Forwarded-For:");
  if (u) {
    u += 16;
    while (*u == ' ')++u;
    a = u;
    for(;;) { if ((*u != '.') && !((*u >= '0') && (*u <= '9'))) break; ++u; }
    *u = 0;
  } else {
    a = getenv("TCPREMOTEIP"); /* IP# of the client */ 
    if (!a) a = "0.0.0.0";
  }

  /* showip-service: just return IP address */
  if (!strncmp(cptr, "showip", 7)) {
    u = strcpy(ur, str200);
    if (a) {
      cptr = strchr(u, '@');
      strcpy(cptr, a);
    }
    write1(u, strlen(u));
    _exit(0);
  }

  /* search DYNDNS list */
  for (i = 0; i < TABLE_LEN; ++i) {
    if (!strncmp(cptr, map_table[i][0], 9)) {
      /* entry found */
      cptr = ur;
      strcpy(cptr, str200);
      u = strchr(cptr, '@');
      strcpy(u, a);
      write1(cptr, strlen(cptr)); /* write HTTP-header */
      u = map_table[i][2];
      p = map_table[i][3];
      unlink(cptr);
      i = open(u, O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);
      if (i > 0) { /* create file for DYNDNS */
#ifdef DEBUG
        strcpy(outbuf, logdynstr);
        strcat(outbuf, cptr);
        strcat(outbuf, ", ");
        strcat(outbuf, a);
        strcat(outbuf, "\n");
        write2(outbuf, strlen(outbuf));
#endif

        strcpy(outbuf, a);
        strcat(outbuf, p);
        write(i, outbuf, strlen(outbuf));
        close(i);

        strcpy(outbuf, logdynstr);
        strcat(outbuf, u);
        strcat(outbuf, "=");
        strcat(outbuf, a);
        strcat(outbuf, "\n");
        write2(outbuf, strlen(outbuf));

#ifdef DEBUG
      } else {
        write1("\nError\n", 7);
#endif
      }
      _exit(0);
    }
  }

  /* short URL feature enabled with symlink or dir? */
  if (chdir("../SHORTURL") == -1) my_exit(8);

  /* is it a short URL? try to read symlink */
  if (cptr[5]) goto noshorturl;
  u = strcpy(ur, str301) + sizeof str301; --u;
  i = readlink(cptr, u, 1024);
  if (i != -1) {
    u += i;
    strcpy(u, "\r\n\r\n");
    u += 4;
    write1(ur, u - ur);
    _exit(0);
  }

  noshorturl:
  if (*cptr != '?') goto no_querystring;
  ++cptr;

  /* show long URL? */
  if (*cptr == 'l') {

    ++cptr;
    if (*cptr != '=') my_exit(8);
    ++cptr;

    /* convert %2F */
    cptr = convert_url(cptr);
  
#ifdef FWB_NULL_DEF

    /* URL is tdyn.de/?u=... */
    len = strlen(cptr);

static jmp_buf check_remote_procstate;

void check_remote_timeout_alarm_sighandler(int num) {
  longjmp(check_remote_procstate,1);
}
    /* set timeout */
    signal(SIGALRM,check_remote_timeout_alarm_sighandler);
    alarm(3);
    if (setjmp(check_remote_procstate)) {
      _exit(9);
    }
  
    p = stpcpy(http_buf, STR_HTTP_GET1);
    p = stpcpy(p, key);
    p = stpcpy(p, STR_HTTP_GET2);
    p = stpcpy(p, ip_address);
    p = stpcpy(p, STR_HTTP_GET3);
    p = http_buf;
    if (inet_aton(ip_address, &server_sockaddr.sin_addr) == 0) _exit(3);
    server_sockaddr.sin_family = AF_INET;
    server_sockaddr.sin_port = htons(port);
  
    if ((sockfd=socket(PF_INET,SOCK_STREAM,0)) == -1) { print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on opening socket\n"); _exit(3); }
    if (tcp_defer_accept) {
      if (setsockopt(sockfd,IPPROTO_TCP,TCP_DEFER_ACCEPT,&tcp_defer_accept,sizeof tcp_defer_accept) != 0) { close(sockfd); print("error on setsockopt"); _exit(3); }
    }
    if (connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr)) == -1) {
      close(sockfd);
      print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on connect\n"); _exit(3);
    }
    if(write(sockfd,p,strlen(p)) == -1) { close(sockfd); print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on write\n"); _exit(3); }
    p = http_buf;
    r = http_buf + 19;
    for(retrycount=0;;) {
      buflen = read(sockfd, p, READBYTES);
      if (buflen == -1) goto err_exit;
      if (buflen > 0) p += buflen;
      if (buflen > 500) {
        /* only do more read()-calls if first read got some hundred bytes */
        int mrb = READBYTES-buflen;
        buflen = read(sockfd, p, mrb);
        while (buflen > 0) {
          mrb = mrb-buflen;
          if (mrb<0) break;
          p += buflen;
          buflen = read(sockfd, p, mrb);
        }
        if (buflen > 0) p += buflen;
      }
      if ((p < r) || buflen) {
        if (retrycount >= MAXRETRIES) {
          err_exit:
          close(sockfd);
          if (buflen == -1) { print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on read\n"); _exit(3); }
          print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - received data underrun #");
          write1(http_buf,p-http_buf);
          print("#\n");
          _exit(3);
        }
        usleep(100);
        retrycount++;
      } else break;
    }
    close(sockfd);
#endif
  }

  /* add short URL? */
  if (*cptr == 'u') {
    int len;
    char shorturl[6];
    crc_t crc = 0;

    ++cptr;
    if (*cptr != '=') my_exit(8);
    ++cptr;

    /* convert %2F */
    cptr = convert_url(cptr);

    /* URL is tdyn.de/?u=... */
    len = strlen(cptr);

    crc = crc_update(crc, (unsigned char *) cptr, len);
    shorturl[5] = 0;
    shorturl[0] = asciitrans(crc >> 24);
    shorturl[1] = asciitrans((crc >> 18) & 0x3f);
    shorturl[2] = asciitrans((crc >> 12) & 0x3f);
    shorturl[3] = asciitrans((crc >> 6) & 0x3f);
    shorturl[4] = asciitrans(crc & 0x3f);

    /* create symlink */
    strcpy(ur, "http://");
    u = ur + 7;
    strcpy(u, cptr);
    u += len;
    unlink(shorturl);
    if (symlink(ur, shorturl) == -1) my_exit(8);

    /* return shortened URL */
    u = stpcpy(ur, HEADER200);
    u = stpcpy(u, "http://tdyn.de/");
    u = stpcpy(u, shorturl);
    write1(ur, u - ur);
    _exit(0);
  }

  no_querystring:
  my_exit(8); /* no entry found/unknown URL: redirect and error */
  _exit(0);
}
